/*The code was written and debugged by Matt Gray, Michael Koppelman,
  and John C. Martin at the University of Minnesota as part of the Eta
  Carinae HST Treasury Project (copyright 2006).  You are free to
  distribute this source as long as this notice remains intact.  End
  users of this code agree to take responsibility for how this code is
  used and the programmers, the Eta Carinae HST Treasury Project, and the
  University of Minnesota bear no liability.*/  

#include "exfits.h"

typedef int rl_icpfunc_t (char*);

typedef struct {
        char *name;
        rl_icpfunc_t *func;
        char *doc;
} COMMAND;

/* Command functions */
int com_open    __P((char *));
int com_info    __P((char *));
int com_log     __P((char *));
int com_trace   __P((char *));
int com_extract __P((char *));
int com_close   __P((char *));
int com_quit    __P((char *));
int com_help    __P((char *));
int execute_line (char *line);
void initialize_readline();
char **exfits_completion(const char *text, int start, int end);
char * command_generator(const char *text, int state);
// extern char *xmalloc();
extern FILE *logfile;

COMMAND commands[] = {
	{"open", com_open, "Open file FILE\n"},
	{"info", com_info, "Display information on the loaded file\n"},
	{"log", com_log, "Change log file to FILE.\n"
"\n"
"\t\tUsage: > log <filename>\n"},
	{"trace", com_trace, "Perform trace on currently open file\n"},
	{"extract", com_extract, "Extract from current file\n"
"\n"
"\t\tThe extract command can be run three different ways. Simply\n"
"\t\ttyping `extract' will cause exfits to prompt for the \n"
"\t\textraction orientation, type, and parameters.\n"
"\t\tAdditionally, you can specify all your choices on a single\n"
"\t\tline, as follows:\n"
"\n"
"\t\t> extract <orientation> <extraction type> <params> <out file>\n"
"\n"
"\t\tNote:  For Echelle extractions, each order is extracted\n"
"\t\t       and outputed to separate files.  Orders without data\n"
"\t\t       are skipped.\n"
"\n"
"\t\tEXAMPLES: \n\n"
"\t\tRow-based square (simple) extraction:\n"
"\t\t> extract row square <min row> <max row> <output file>\n\n"
"\t\tRow-based mesa extraction:\n"
"\t\t> extract row mesa <center> <base> <peak> <output file>\n\n"
"\t\tColumnar square extraction:\n"
"\t\t> extract col square <min col> <max col> <output file>\n\n"
"\t\tEchelle square extraction:\n"
"\t\t> extract ech square <min col> <max col> <output file>\n\n"
"\t\tEchelle mesa extraction:\n"
"\t\t> extract ech mesa <center> <base> <peak> <output_file>\n\n"
"\t\tEchelle complete extraction:\n"
"\t\t> extract ech all <output_file>\n"},
	{"close", com_close, "Close current file\n"},
	{"quit", com_quit, "Quit program\n"},
	{"help", com_help, "Display this text\n"},
	{"?", com_help, "Synonym for `help'\n"},
	{ (char *)NULL, (rl_icpfunc_t *)NULL, (char *)NULL}
};

COMMAND *find_command ();

/* Global variables */
int done = 0;
struct fitsdata * fits;
struct trace trc;
extern struct arguments arguments;
static char *row_text = "row";
static char *col_text = "column";

/* Look up NAME as the name of a command, and return a pointer to the
 * command. NULL if the command is not a command name */
COMMAND *
find_command (char * name)
{
	int i;

	for (i = 0; commands[i].name; i++)
		{
			if (strcmp(name,commands[i].name) == 0)
				{
					return (&commands[i]);
				}
		}

	return ((COMMAND *)NULL);
}

char *
dupstr (char *s)
{
	char *r;

	r = (char *)malloc (sizeof(char)*(strlen (s) + 1));
	strcpy (r, s);
	return (r);
}


char * stripwhite (char *string)
{
	register char *s, *t;

	for (s = string; *s==' '; s++)
		;
	
	if (*s == 0)
		return (s);
	
	t = s + strlen (s) - 1;
	while (t > s && (*t==' '))
		t--;
	*++t = '\0';

	return s;
}
		

int
interactive_mode ()
{
	char *line = NULL;
	char *s    = NULL;

	printf ("exfits v%04.2f\n", EXFITS_VERSION);
	printf ("Created at the University of Minnesota.\n");
	printf ("(c)2006, Matt Gray & John C. Martin.\n");
	printf ("\n");
	
	initialize_readline ();

	while (done == 0)
		{
			/* get the next command */
			line = readline ("exfits> ");

			if (! line)
				break;

			/* Remove leading, trailing whitespace */
			s = stripwhite(line);

			if (*s)
				{
					add_history(s);
					execute_line(s);
				}

			free(line);
			
		}
	return 0;
}

/* Execute a command line */
int
execute_line (char *line)
{
	int i;
	COMMAND *command;
	char *word;

	/* Isolate the comand word. */
	i = 0;
	while (line[i] && (line[i]==' '))
		i++;
	
	word = line + i;
	
	while (line[i] && !(line[i]==' '))
		i++;

	if (line[i])
		line[i++] = '\0';
	
	command = find_command(word);

	if (!command) 
		{
			fprintf (stderr,  "ERROR: %s: No such command.\n", word);
			fprintf (logfile, "[%s] ERROR: %s: No such command.\n", logfile_time(), word);
			return 1;
		}
	
	/* Get argument to command */
	while (line[i]==' ')
		i++;

	word = line + i;

	/* call the func */
	return ((*command->func) (word));
}
	
void
initialize_readline ()
{
  rl_readline_name = "exfits";
	rl_attempted_completion_function = exfits_completion;
}

char **
exfits_completion (const char *text, int start, int end)
{
	char **matches;

	matches = (char **) NULL;

	if (start == 0)
		matches = completion_matches(text,command_generator);
	
	return (matches);
}

char *
command_generator (const char *text, int state)
{
  static int list_index, len;
  char *name;

  /* If this is a new word to complete, initialize now.  This
     includes saving the length of TEXT for efficiency, and
     initializing the index variable to 0. */
  if (!state)
    {
      list_index = 0;
      len = strlen (text);
    }

  /* Return the next name which partially matches from the
     command list. */
  while ((name = commands[list_index].name))
    {
      list_index++;

      if (strncmp (name, text, len) == 0)
        return (dupstr(name));
    }

  /* If no names matched, then return NULL. */
  return ((char *)NULL);
}

int com_info (char *arg)
{
	if (!fits) {
	  fprintf (stderr, "ERROR: No file opened.\n");
	  printf ("Please use the `open' command to open a file.\n");
	  return 1;
	}
	printf("File info:\n");

	printf(" Rows: %i\n Cols: %i\n",fits->rows,fits->cols);
	if (fits->pech!=0) printf(" Echelle Order Sep: %i\n",fits->ordsep);
	printf(" CRPIX1: %g\n CRPIX2: %g\n",fits->crpix1,fits->crpix2);
	printf(" CRVAL1: %g\n CRVAL2: %g\n",fits->crval1,fits->crval2);
	printf(" CD1_1 : %g\n CD1_2 : %g\n",fits->cd1_1,fits->cd1_2);
	printf(" CD2_1 : %g\n CD2_2 : %g\n",fits->cd2_1,fits->cd2_2);
	printf(" PIXVAL1: %g\n PIXVAL2: %g\n",fits->pixval1,fits->pixval2);

	return 0;
  
}

int com_open (char *arg)
{
	FILE *fp;
	if (!*arg)
		{
			fprintf (stderr, "ERROR: no file specified.\n");
			fprintf (stderr, "Usage: open <filename>\n");
			return 1;
		}

	/* Attempt to open the file */
	fp = fopen (arg, "r");
	/*
	if (!fp)
		{
			fprintf (stderr, "ERROR: %s: file does not exist.\n", arg);
			fprintf (logfile, "[%s] ERROR: %s: file does not exist.\n", logfile_time(), arg);
			return 1;
		}
	*/
	
	/* Load up FITS */
	if (fits)
		{
			free (fits);
		}
	fits = fitsdata_read(arg);
	if (!fits)
		{
			fprintf (stderr, "ERROR: Could not read FITS format.\n");
			fprintf (logfile, "[%s] ERROR: Could not read FITS format.\n", logfile_time());
			return 1;
		}
	
	/* Set current file name */
	arguments.input_file = (char *) calloc (strlen (arg) + 1, sizeof (char));
	strcpy (arguments.input_file, arg);

	fprintf (stderr, "INFO : Opened file %s.\n", arguments.input_file);
	fprintf (logfile, "[%s] INFO : Opened file %s.\n", logfile_time (), arguments.input_file);
	
	return 0;
}

int com_log (char *arg)
{
	FILE *fp;

	if (!*arg)
		{
			fprintf (stderr, "ERROR: log: No log file specified.\n");
			fprintf (logfile, "[%s] ERROR: log: No log file specified.\n", logfile_time());
			return 1;
		}
	
	/* Attempt to open the file */
	fp = fopen (arg, "a");

	if (!fp)
		{
			fprintf (stderr, "ERROR: log: Unable to open file %s\n", arg);
			fprintf (logfile, "[%s] ERROR: log: Unable to open file %s\n", logfile_time(), arg);
			return 1;
		}
	
	fprintf (stderr, "INFO : Logfile switched to `%s' at user request.\n", arg);
	fprintf (logfile, "[%s] INFO : Logfile switched to `%s' at user request.\n", logfile_time(), arg);
	
	logfile = fp;

	fprintf (logfile, "[%s] INFO : Log file opened from interactive session.\n", logfile_time());

	return 0;
}

int com_trace (char *arg)
{
	double guess;

	if (!fits)
		{
			fprintf (stderr, "ERROR: No file opened.\n");
			printf ("Please use the `open' command to open a file.\n");
			return 1;
		}

	if (fits->pech) {
	  fprintf(stderr,"ERROR:  This is Echelle data.  Tracing it is futile\n");
	  return 1;
	}
	fitsdata_tracenfit (fits, &guess, &trc, stdout);
	printf ("INFO : Central row guess: %lf\n", guess + 1);
	fprintf (logfile, "[%s] INFO : Central row guess: %lf\n", logfile_time(), guess + 1.);
	printf ("INFO : Parameters for trace fit: %g, %g, %g, %g, %g\n",trc.param[1],trc.param[2],trc.param[3],trc.param[4],trc.param[5]);
	fprintf (logfile,"[%s] INFO : Parameters for trace fit: %g, %g, %g, %g, %g\n",logfile_time(),trc.param[1],trc.param[2],trc.param[3],trc.param[4],trc.param[5]);
	return 0;
}

int com_extract (char *arg)
{
	char *line;
        int ptrace=0;
	int line_len = 0;
	/* Size of unit (row,col) */
	int bound = 0;
	/* Point at {row,column} text static vars */
	char *type_text = NULL;

	/* Make sure we have a FITS file */
	if (!fits)
		{
			fprintf (stderr, "ERROR: No file currently opened.\n");
			printf ("Use the `open' command to load a FITS file before extracting.\n");
			return 1;
		}

	/* Was the extraction type specified? */
	if (!*arg)
		{
			FILE *fp;
			/* The case where no type is specified */
			int choice = 0;
			/* Square params */
			int start, end;
			/* Mesa params */
			double basew, platw, cenrow;
			/* Track input loops */
			int done  = 0;

			/* Prompt for extraction type */
			printf ("Available extractions:\n");
			printf ("  Row     1. Square\n");
			printf ("          2. Mesa\n\n");
			printf ("  Col     3. Square\n");
			if (fits->pech) {
			  printf ("\n  Echelle 4. Square\n");
			  printf ("          5. Mesa\n");
			  printf ("          6. All\n");
			}
			while (choice == 0)
				{
					if (fits->pech) line = readline ("[1-6]: ");
					else line = readline("[1-3]: ");
					if (!*line || (sscanf(line, "%d", &choice) != 1) ||
						(choice < 1) || (choice > 3))
						{
						        if ((fits->pech)&&(choice < 7)) free(line);
						        else {
							  fprintf (stderr, "ERROR: Invalid choice `%s'.\n", line);
							  if (line)
							    free (line);
							  choice = 0;
							}
						}
					else
						free (line);
				}

			/* We have a valid choice */
			arguments.type = choice;
			if (fits->pech) {
			  printf("An Echelle extraction means that the same rows will be\n");
			  printf("extracted from each order.  The first valid row in an order \n");
			  printf("is 1 and the last is %i.\n",fits->ordsep);
			}

			switch (choice)
				{
					/* Handle the square extractions */
					case 1:
						/* Rows */
						type_text = row_text;
						bound     = fits->rows;
					case 3:
						/* Cols */
						if (!type_text)
							{
								type_text = col_text;
								bound     = fits->cols;
							}
						printf ("Enter the start and end %ss for the extraction, using FITS-style numbering.\n", type_text);
						printf ("Input should be two integers separated by a space.\n");
						printf ("<start %s> <end %s>\n", type_text, type_text);
						while (!done)
							{
								line = readline ("> ");
								if (!*line || (sscanf(line, "%d %d", &start, &end) != 2))
									{
										fprintf (stderr, "ERROR: Invalid input `%s'.\n", line);
										printf  ("Input should be two integers separated by a space.\n");
										printf  ("<start %s> <end %s>\n", type_text, type_text);
										if (line)
											free (line);
									}
								else
									{
										free (line);

										/* Check for transposition */
										if (end < start)
											{
												printf ("info: End %s smaller than start %s. Swapping...\n", type_text, type_text);
												int tmp;
												tmp = end;
												end = start;
												start = tmp;
											}
											
										/* Check for bounds of FITS file */ 
										if (start < 1)
											{
												fprintf (stderr, "ERROR: Start %s too small for FITS file.\n", type_text);
												printf ("Try a value in the range [1,%d]\n", bound);
											}
										else if (end > bound)
											{
												fprintf (stderr, "ERROR: End %s too large for FITS file.\n", type_text);
												printf ("Try a value in the range [1.%d]\n", bound);
											}
										else
											{
												/* Valid parameters. */
											  arguments.minrow = start; 
											  arguments.maxrow = end;   
											  done = 1;
											}
										
									}
							}
						break;
					case 4:
						/* Ech Cols */
					        type_text = row_text;
						bound     = fits->ordsep;
						printf ("Enter the start and end %ss for the extraction, using FITS-style numbering.\n", type_text);
						printf ("Input should be two integers separated by a space.\n");
						printf ("<start %s> <end %s>\n", type_text, type_text);
						while (!done)
							{
								line = readline ("> ");
								if (!*line || (sscanf(line, "%d %d", &start, &end) != 2))
									{
										fprintf (stderr, "ERROR: Invalid input `%s'.\n", line);
										printf  ("Input should be two integers separated by a space.\n");
										printf  ("<start %s> <end %s>\n", type_text, type_text);
										if (line)
											free (line);
									}
								else
									{
										free (line);

										/* Check for transposition */
										if (end < start)
											{
												printf ("info: End %s smaller than start %s. Swapping...\n", type_text, type_text);
												int tmp;
												tmp = end;
												end = start;
												start = tmp;
											}
											
										/* Check for bounds of FITS file */ 
										if (start < 1)
											{
												fprintf (stderr, "ERROR: Start %s too small for FITS file.\n", type_text);
												printf ("Try a value in the range [1,%d]\n", bound);
											}
										else if (end > bound)
											{
												fprintf (stderr, "ERROR: End %s too large for FITS file.\n", type_text);
												printf ("Try a value in the range [1.%d]\n", bound);
											}
										else
											{
												/* Valid parameters. */
											  arguments.minrow = start; 
											  arguments.maxrow = end;   
											  done = 1;
											}
										
									}
							}
						break;
					case 2:
						/* Mesa extraction over rows */
						type_text = row_text;
						bound     = fits->rows;
						done = 0;
						printf ("Enter the parameters for mesa extraction.\n");
						printf ("All parameters may be floating-point numbers.\n");
                                                while ((done<1)||(done>2)) {
                                                  printf ("Do you want to: \n   (1) Follow a trace\n   (2) Give a central row\n[1-2]: ");
                                                  scanf("%i",&done);
                                                }
                                                if (done == 1) {
                                                  /*Follow the trace*/
						  ptrace=1;
                                                  done=0;
						  printf ("<row offset from trace> <base width> <peak width>\n");
						  while (!done)
							{
								line = readline("> ");
								if (!*line || (sscanf(line, "%lf %lf %lf", &cenrow, &basew, &platw) != 3))
									{
										fprintf (stderr, "ERROR: Invalid input `%s'.\n", line);
										printf ("Input should be a space-separated list of floating point numbers.\n");
										printf ("<row offset from trace> <base width> <peak width>\n");
										if (line)
											free (line);
									}
								else
									{
										free (line);

										/* Check for transposition */
										if (platw > basew)
											{
												printf ("info: Peak width larger than base width. Swapping...\n");
												double tmp;
												tmp = basew;
												basew = platw;
												platw = tmp;
											}
										/* Make sure central row is in-bounds */
										if (trc.param[1] - cenrow < 1 || cenrow + trc.param[1] > bound)
											{
												fprintf (stderr, "ERROR: Central %s not within FITS file.\n", type_text);
												printf ("Try a value in the range [1,%d]\n", bound);
											}
										else
											{
	 											/* Valid parameters. */
                                                                                                trc.roff=cenrow;
 											        arguments.cenrow = cenrow;
												arguments.platw  = platw;
												arguments.basew  = basew;
												done = 1;
											}
									}
							  }

                                                } else {
                                                  /*Enter a central row*/
                                                  done=0;
						  printf ("<central %s> <base width> <peak width>\n", type_text);
						  while (!done)
							{
								line = readline("> ");
								if (!*line || (sscanf(line, "%lf %lf %lf", &cenrow, &basew, &platw) != 3))
									{
										fprintf (stderr, "ERROR: Invalid input `%s'.\n", line);
										printf ("Input should be a space-separated list of floating point numbers.\n");
										printf ("<central %s> <base width> <peak width>\n", type_text);
										if (line)
											free (line);
									}
								else
									{
										free (line);

										/* Check for transposition */
										if (platw > basew)
											{
												printf ("info: Peak width larger than base width. Swapping...\n");
												double tmp;
												tmp = basew;
												basew = platw;
												platw = tmp;
											}
										/* Make sure central row is in-bounds */
										if (cenrow < 1 || cenrow > bound)
											{
												fprintf (stderr, "ERROR: Central %s not within FITS file.\n", type_text);
												printf ("Try a value in the range [1,%d]\n", bound);
											}
										else
											{
	 											/* Valid parameters. */
 											        arguments.cenrow = cenrow;
												arguments.platw  = platw;
												arguments.basew  = basew;
												done = 1;
											}
									}
							  }
                                                        }
						break;
					case 5:
						/* Echelle Mesa extraction */
						type_text = row_text;
						bound     = fits->ordsep;
						done = 0;
						printf ("Enter the parameters for mesa extraction.\n");
						printf ("All parameters may be floating-point numbers.\n");
						printf ("<central %s> <base width> <peak width>\n", type_text);
						while (!done)
							{
								line = readline("> ");
								if (!*line || (sscanf(line, "%lf %lf %lf", &cenrow, &basew, &platw) != 3))
									{
										fprintf (stderr, "ERROR: Invalid input `%s'.\n", line);
										printf ("Input should be a space-separated list of floating point numbers.\n");
										printf ("<central %s> <base width> <peak width>\n", type_text);
										if (line)
											free (line);
									}
								else
									{
										free (line);

										/* Check for transposition */
										if (platw > basew)
											{
												printf ("info: Peak width larger than base width. Swapping...\n");
												double tmp;
												tmp = basew;
												basew = platw;
												platw = tmp;
											}
										/* Make sure central row is in-bounds */
										if (cenrow < 1 || cenrow > bound)
											{
												fprintf (stderr, "ERROR: Central %s not within FITS file.\n", type_text);
												printf ("Try a value in the range [1,%d]\n", bound);
											}
										else
											{
	 											/* Valid parameters. */
 											        arguments.cenrow = cenrow;
												arguments.platw  = platw;
												arguments.basew  = basew;
												done = 1;
											}
									}
							}
						break;
				        case 6:
 				                /*Echelle extract all*/
						type_text = row_text;
						bound     = fits->ordsep-1;
						printf("Extracting all %ss in each order.\n",type_text);
						arguments.minrow = 1; 
						arguments.maxrow = bound;   
						break;
					default:
						/* Should never get here */
						fprintf (stderr, "ERROR: Invalid choice of extraction? choice=%d\n", choice);
						return 1;
				} /* end switch */

			/* Prompt for file name */
			printf ("Please enter an output file name: \n");
			line = readline(">");

			line_len = strlen(line);

			/* File extension checking */
			if ((line_len > 5) && strncmp( line+(line_len-5), ".fits", 5 ) == 0)
			{
				fprintf( stderr, "INFO : Outputting to FITS\n" );
				arguments.fits = 1;
			}
			else if ((line_len > 4) && strncmp( line+(line_len-4), ".uni", 3 ) == 0)
			{
				fprintf( stderr, "INFO : Outputting to unifspec\n" );
				arguments.unifspec = 1;
			}
			else if (!arguments.fits && !arguments.unifspec)
			{
				fp = fopen(line, "r");

				while (fp)
					{
						fprintf(stderr, "ERROR: File exists: %s\n",line);
						fprintf(logfile, "[%s] ERROR: File exists: %s\n", logfile_time(), line);
						fclose(fp);
						free(line);
						printf ("Please enter an output file name: \n");
						line = readline(">");
						fp = fopen(line, "r");
					}
			}
				
			arguments.output_file = (char *) calloc (strlen(line) + 1, sizeof(char));
			strcpy (arguments.output_file, line);
			free(line);

			/* We have valid params for whatever extraction we are doing. */
		}
	else //arguements have been specified and this will not be interacitive
		{
			char *head, *tail;
			char filebuffer[1024];

			/* Arguments were specified. */

			/* Interactively determine the arguments */

			/* Find first word */
			for (head = arg; *head && (*head==' '); head++)
				;

			/* Missing first arg */
			if (!*(head+1))
				{
					fprintf (stderr, "ERROR: extract: missing orientation (one of {row, col})\n");
					return 1;
				}
			
			for (tail = head; *tail && !(*tail==' '); tail++)
				;
			*tail = '\0';

			/* Discovering extraction orientation */
			if (strcmp(head, "row") == 0)
				{
					/* row-based extraction */
					arguments.type = 0; /* Add to					arguments.minrow = start;
					arguments.maxrow = end;
 this later when we choose type */
					bound     = fits->rows;
					type_text = row_text;
				}
			else if (strcmp(head, "col") == 0)
				{
					/* col-based extraction */
					arguments.type = 2;  /* Add to this later when we choose type */
					bound     = fits->cols;
					type_text = col_text;
				}
			else if (strcmp(head, "ech") == 0)
				{
					/* Echelle extraction */
					arguments.type = 3;  /* Add to this later when we choose type */
					bound     = fits->ordsep;
					type_text = row_text;
				}
			else 
				{
					/* Incorrect format */
					fprintf (stderr, "ERROR: extract: Incorrect argument `%s'\n", head);
					return 1;
				}

			/* Get next word */
			for (head = ++tail; *head && (*head==' '); head++)
				;

			if (!*(head+1))
				{
					fprintf (stderr, "ERROR: extract: Missing extraction type (one of {square,mesa})\n");
					return 1;
				}

			for (tail = head; *tail && !(*tail==' '); tail++)
				;
			*tail = '\0';

			/* Extraction type block */
			if (strcmp(head, "square") == 0)
				{
					int start, end;
					int scanned;
					/* Square extraction */
					arguments.type += 1;  /* If row-based, ends up at 1, col-based at 3, ech-based at 4 */

					scanned = sscanf (++tail, "%d %d", &start, &end);
					if (scanned != 2)
						{
							fprintf (stderr, "ERROR: extract: Incorrect number of %s parameters.\n", type_text);
							fprintf (stderr, "The format is:\n");
							fprintf (stderr, "> extract %3s square <start %s> <end %s> <output file>\n", type_text, type_text, type_text);
							return 1;
						}
					/* Sanity checks */
					if (start > end)
						{
							int tmp;
							/* Transposed */
							printf ("INFO : Start %s larger than end %s. Swapping...\n", type_text, type_text);
							tmp = start;
							start = end;
							end   = tmp;
						}
					if (start < 1 || start > bound)
						{
							fprintf (stderr, "ERROR: start %s is outside the FITS boundaries.\n", type_text);
							fprintf (stderr, "       Try choosing a %s within [1,%d]\n", type_text, bound);
							return 1;
						}
					if (end > bound)
						{
							fprintf (stderr, "ERROR: end %s is outside the FITS boundaries.\n", type_text);
							fprintf (stderr, "       Try choosing a %s within [%d,%d]\n", type_text, start, bound);
							return 1;
						}
					arguments.minrow = start;
					arguments.maxrow = end;

					if (sscanf(tail, "%d %d %s", &start, &end, filebuffer) != 3)
						{
							*filebuffer = '\0';
						}

					/* Params are valid at this point. */
				}
			else if (strcmp(head, "mesa") == 0)
				{
					double cenrow, platw, basew;
					int scanned;

					/* Mesa extraction */
					arguments.type += 2;  /* Becomes 2 for row-based, 4 for col-based (generates an error), 5 for ech-based */
					if (arguments.type == 4)
						{
							fprintf (stderr, "ERROR: Mesa extraction not currently supported for columns.\n");
							fprintf (logfile, "[%s] ERROR: Mesa extraction not currently supported for columns.\n", logfile_time());
							return 1;
						}

					scanned = sscanf (++tail, "%lf %lf %lf", &cenrow, &basew, &platw);

					if (scanned != 3)
						{
							fprintf (stderr, "ERROR: Incorrect number of parameters for type `mesa'\n");
							fprintf (stderr, "       The format is:\n");
							fprintf (stderr, "       > extract %3s mesa <central row> <base width> <peak width> <output file>\n", type_text);
							return 1;
						}
					/* Sanity checks */
					if (cenrow < 1.0 || cenrow > bound)
						{
							fprintf (stderr, "ERROR: Central %s not within FITS file.\n", type_text);
							fprintf (stderr, "       Try a value in [1,%d]\n", bound);
							return 1;
						}
					if (platw > basew)
						{
							double tmp;
							fprintf (stderr, "INFO : Peak width larger than base width. Swapping...\n");
							fprintf (logfile, "[%s] INFO : Peak width larger than base width. Swapping...\n", logfile_time());
							tmp   = basew;
							basew = platw;
							platw = tmp;
						}

					if (sscanf(tail, "%lf %lf %lf %s", &cenrow, &basew, &platw, filebuffer) != 4)
						{
							*filebuffer = '\0';
						}

					/* Place cenrow in zero-order */
					arguments.cenrow = cenrow;
					arguments.basew  = basew;
					arguments.platw  = platw;
					
					/* Valid parameters*/
				}
			else if (strcmp(head, "all") == 0)
				{
					arguments.type += 3;  /* Becomes 3 for row-based, 5 for col-based (generates an error), 6 for ech-based */
					if (arguments.type < 6)
						{
							fprintf (stderr, "ERROR: Extraction of all rows only supported for Echelle extractions.\n");
							fprintf (logfile, "[%s] ERROR: Extraction of all rows only supported for Echelle extractions.\n", logfile_time());
							return 1;
						}
					arguments.minrow = 1;
					arguments.maxrow = bound;
				}
			else
				{
					fprintf (stderr, "ERROR: extract: Incorrect argument `%s'\n", head);
					return 1;
				} // End extraction type block


			if (!*filebuffer)
				{
					fprintf (stderr, "ERROR: No output file specified.\n");
					return 1;
				}

			if (strcmp ("-", filebuffer) == 0 || strcmp("stdout", filebuffer) == 0) 
				{
					/* We are writing to standard out */
					fprintf (stderr, "INFO : Writing to standard output.\n");
					fprintf (logfile, "[%s] INFO : Writing to standard output.\n", logfile_time());
					arguments.output_file = NULL;
				}
			else
				{
					/* Attempt to open the file */
					FILE *fp;
					fp = fopen (filebuffer, "r");
					if (fp)
						{
							fprintf (stderr, "ERROR: extract: File `%s' already exists.\n", filebuffer);
							fclose (fp);
							return 1;
						}
					// arguments.output_file = (char *) calloc (strlen(head) + 1, sizeof (char));
					// strcpy (arguments.output_file, head);
					arguments.output_file = filebuffer;

				}
		}

	/* Perform the extraction */
	arguments.cenrow--; // decriment these by one because row 0 in the array = row 1 in reality
	arguments.minrow--;
	arguments.maxrow--;
	do_extract(fits, &trc, ptrace, &arguments);
	fprintf (stderr, "INFO : Extraction complete.\n");
	fprintf (logfile, "[%s] INFO : Extraction complete.\n", logfile_time());

	return 0;
}

int com_close (char *arg)
{
	if (fits)
		{
			free (fits);
		}
	arguments.input_file = NULL;
	printf ("status: File closed.\n");
	return 0;
}

int com_quit (char *arg)
{
	if (fits)
		{
			free (fits);
		}
	done = 1;
	return 0;
}

int com_help (char *arg)
{
	int i;
	int printed = 0;

	for (i = 0; commands[i].name; i++)
		{
			if (!*arg || (strcmp (arg, commands[i].name) == 0))
				{
					printf ("%s\t\t%s\n", commands[i].name, commands[i].doc);
					printed++;
				}
		}

	if (!printed)
		{
			printf ("No commands match `%s'.	Possibilties are:\n", arg);

			for (i = 0; commands[i].name; i++)
				{
					/* Print in six columns. */
					if (printed == 6)
						{
							printed = 0;
							printf ("\n");
						}

					printf ("%s\t", commands[i].name);
					printed++;
				}

			if (printed)
				printf ("\n");
		}
	return 0;
}
