/*The code was written and debugged by Matt Gray, Michael Koppelman,
  and John C. Martin at the University of Minnesota as part of the Eta
  Carinae HST Treasury Project (copyright 2006).  You are free to
  distribute this source as long as this notice remains intact.  End
  users of this code agree to take responsibility for how this code is
  used and the programmers, the Eta Carinae HST Treasury Project, and the
  University of Minnesota bear no liability.*/  

/* main.c
* 
* main file for exfits, driver program, etc 
*/

#include "exfits.h"

const char *argp_program_version = "exfits 2.0";
const char *argp_program_bug_address = "<webmaster@aps.umn.edu>";

static char doc[] =
  "\n`exfits' is a program designed to perform extractions on HST STIS \n"
  "2D spectra from FITS files.\n"
  "\n"
  "exfits has two primary modes: batch and interactive. When run without "
  "arguments, or when using the `-i' switch, exfits will be run "
  "interactively.\n"
  "\n"
  "Otherwise, the program can be run with command-line input only. There "
  "are currently six extraction types:\n"
  "   `-c'    Mesa-shaped extraction across floating-point rows\n"
  "   `-s'    Box-shaped extraction across integer rows\n"
  "   `-t'    Box-shaped vertical extraction with integer rows\n"
  "   `-e -c' Echelle mode mesa-shaped extraction across floating-point rows\n"
  "   `-e -s' Echelle mode box-shaped extraction across integer rows\n"
  "   `-e'    Echelle mode extraction of all rows\n"
  "\n"
  "Properly formated 2D Echelle spectra (which have the ORDERSEP keyword\n"
  "in their flux HDU) can be extracted in either normal or Echelle mode\n"
  "('-e').  In Echelle mode the row numbers are treated relative to each \n"
  "order and each order is written to a separate file.\n"
  "\n"
  "The extraction types including '-c' can take up to three parameters. \n"
  "For a non-Echelle FITS file, omitting the central row causes the program\n"
  "to perform a trace which automatically chooses the brightest row in the\n"
  "file for the center.\n"
  "\n"
  "exfits records its doings in a log file, which defaults to `exfits.log' \n"
  "in the current working directory. exfits will also write \n"
  "cspec-compatible text files, with appropriate headers commented with \n"
  "the `#' character.\n";
static char args_doc[] = "[FILE]";
static struct argp_option options[] = {
  {"verbose", 'v', 0, 0,
   "Display info about extraction, default if interactive."},
  {0, 'o', "FILE", 0, "Output extraction data to FILE. Default is stdout."},
  {0, 'l', "FILE", 0,
   "File to write session log to. Default is `exfits.log'"},
  {0, 's', "MINROW,MAXROW", 0,
   "Row bounds (integer) for simple extraction (inclusive)."},
  {0, 'c', "PLATW,BASEW[,CENTER]", 0,
   "Extract with careful method, specify mesa plateau, base, and optionally row center. If the center is not specified, a trace will determine the center."},
  {"interactive", 'i', 0, 0, "Enter interactive terminal session."},
  {"unifspec", 'u', 0, 0, "Output unifspec instead of the default text."},
  {"fits", 'f', 0, 0, "Output FITS image."},
  {0, 't', "MINCOL,MAXCOL", 0,
   "Extract vertically, with an integer range of columns."},
  {0,'e',0,0,"Extract data in Echelle mode.  (Fails if the input FITS are not properly formated Echelle data)"},
  {0}
};

static error_t parse_opt (int key, char *arg, struct argp_state *state);
static struct argp argp = { options, parse_opt, args_doc, doc };
struct arguments arguments;

FILE *logfile;

int
main (int argc, char **argv)
{

  struct fitsdata *fits;
  struct trace *trc;

	FILE *fp;

/* Default values for arguments */
  arguments.verbose = 0;
  arguments.output_file = NULL;
  arguments.input_file = "-";
  arguments.logfile = "exfits.log";
  arguments.echelle = 0;
  arguments.unifspec = 0;
  arguments.fits     = 0;
  arguments.interactive = 0;
  arguments.type = 0;
  arguments.guess_center = 1;

/* Paranoia */
  arguments.minrow = 0;
  arguments.maxrow = 0;
  arguments.cenrow = 0.0;
  arguments.platw = 0.0;
  arguments.basew = 0.0;

  argp_parse (&argp, argc, argv, 0, 0, &arguments);

/* Attempt to open logfile. */
  logfile = fopen (arguments.logfile, "a");
  if (!logfile)
    {
      fprintf (stderr, "ERROR: Unable to open logfile `%s' in current"
               " working directory. Check permissions, or specify a"
               " logfile location with the -l option.\n", arguments.logfile);
      exit (1);
    }

	if (arguments.fits && arguments.unifspec)
	{
		fprintf( stderr, "ERROR: Unable to write _both_ a FITS file AND a unifspec file!\n" );
		exit (1);
	}

/* Run interactive mode */
	if (arguments.interactive == 1 || ((arguments.type == 0)&&(!arguments.echelle)))
    {
			int retval;
			fprintf (logfile, "[%s] ======== exfits interactive session started. ========\n", logfile_time ());
      retval = interactive_mode (logfile);
			if (retval)
				{
					fprintf (logfile, "[%s] ERROR: interactive_mode returned %d\n", logfile_time(), retval);
				}
			fprintf (logfile, "[%s] ======== exfits interactive session ended.   ========\n", logfile_time ());
			fclose (logfile);
			exit (retval);
    }

/* We are officially in batch mode */

/* Logfile is open, print out some session info */
	fprintf (logfile, "[%s] ======== exfits batch session started.       ========\n", logfile_time ());

/* Check to see if file exists. */
  if (arguments.input_file[0] != '-')
    {
/* We're not opening stdin */
      fp = fopen (arguments.input_file, "r");
      if (!fp)
        {
          fprintf (stderr, "ERROR: Unable to open input file `%s'.\n",
                   arguments.input_file);
          fprintf (logfile, "[%s] ERROR: Unable to open input file `%s'.\n",
                   logfile_time (), arguments.input_file);
          exit (1);
        }
			fclose(fp);
    }
  else
    {
      fprintf (logfile, "[%s] Reading from stdin.\n", logfile_time ());
    }

/* We have the logfile open and verified that the input file is readable. */

  fits = fitsdata_read (arguments.input_file);
  if ((!fits->pech) && (arguments.echelle)) { /*Can't treat a file that is not Echelle as Echelle*/
    fprintf(stderr,"ERROR: Can't use Echelle mode on a non-Echelle FITS file\n");
    fprintf(logfile,"[%s] ERROR: Can't use Echelle mode on a non-Echelle FITS file\n",logfile_time ());
    return(1);
  }

  if ((arguments.echelle)&&(arguments.type!=3)) {
    while (arguments.type<4) arguments.type+=3;
  }

  if (fits == NULL)
    {

      fprintf (stderr, "ERROR: Problem reading FITS file %s\n",
               arguments.input_file);
      fprintf (logfile, "[%s] ERROR: Problem reading FITS file %s\n",
               logfile_time (), arguments.input_file);
    }
  else
    {
      if (arguments.verbose)
        fprintf (stderr, "INFO : Opened file %s.\n", arguments.input_file);
      fprintf (logfile, "[%s] INFO : Opened file %s.\n", logfile_time (), arguments.input_file);
    }

  if ((arguments.type == 2) || (arguments.type == 4))
    {
/* We have all the requisite parameters decided. */
      if (arguments.platw > arguments.basew)
        {
/* We have an inversion of parameters; warn the user and fix */
          double temp;
          fprintf (stderr,
                   "WARN : Peak width larger than base width! Switching...\n");
          fprintf (logfile,
                   "[%s] WARN : Peak width larger than base width! Switching...\n",
                   logfile_time ());
          temp = arguments.platw;
          arguments.platw = arguments.basew;
          arguments.basew = temp;
        }


      if ((arguments.guess_center == 1)&&(!arguments.echelle))
        {
          if (arguments.verbose)
            {
              fprintf (stderr, "INFO : Performing trace..\n");
              fitsdata_trace (fits, &(arguments.cenrow), stdout);
            }
          fprintf (logfile, "[%s] INFO : Performing trace..\n",
                   logfile_time ());
          fitsdata_trace (fits, &(arguments.cenrow), logfile);
        }
      else
        {
/* Subtract one from user-supplied row to get it into 0-order */
	  if (arguments.cenrow <= 0)
	    {
	      if (arguments.verbose)
		{
		  fprintf (stderr, "ERROR : A central row must be given for this extraction.\n");
		}
	      fprintf (logfile, "[%s] ERROR : A central row must be given for this extraction.\n",logfile_time ());
	      return(1);
	    }
	     
          arguments.cenrow -= 1.0;
	}
    }
  else if (arguments.type==6) 
    {
      /*Then do an Echelle all extraction*/
      if (arguments.verbose)
        {
          fprintf (stderr, "Performing an Echelle extraction of all rows for each order.\n");
        }
      fprintf (logfile, "[%s] Performing an Echelle extraction of all rows for each order.\n",
               logfile_time ());
    }
  else                                /* Type must be 1, 3, or 5 here */
    {
      if (arguments.verbose)
        {
          fprintf (stderr, "Simple extraction selected.\n");
          fprintf (stderr, "Min row, max row: %d, %d\n", arguments.minrow,
                   arguments.maxrow);
        }
      fprintf (logfile, "[%s] Simple extraction selected.\n",
               logfile_time ());
      fprintf (logfile, "[%s] Min row, max row: %d, %d\n", logfile_time (),
               arguments.minrow, arguments.maxrow);
    }

  do_extract( fits, trc, 0, &arguments);

  if (arguments.verbose)
    {
      fprintf (stderr, "Wrote output to file `%s'.\n", arguments.output_file);
    }
  fprintf (logfile, "[%s] Wrote output to file `%s'.\n", logfile_time (),
           arguments.output_file);
  fitsdata_destroy (fits);

	fprintf (logfile, "[%s] ======== exfits batch session ended.         ========\n", logfile_time ());

  fclose (logfile);

  return 0;
}

int 
write_header (FILE *fp, struct arguments *args, int order, int usetrace)
{
/* Write out a header */

					
  fprintf (fp, "# exfits v%04.2f extraction\n", EXFITS_VERSION);
  fprintf (fp, "# \n");
  fprintf (fp, "# date         : %s\n", logfile_time ());
  fprintf (fp, "# source       : %s\n", args->input_file);
  if (args->type > 3) 
    fprintf (fp, "# echelle order %i\n",order);
  if ((args->type == 2) || (args->type == 5))
    {
      fprintf (fp, "# type      : mesa\n");
			/* Put it in 1-order row numbering */
      if (usetrace) fprintf(fp,"# Peak trace used.\n# Extract center offset from peak:  %f\n",args->cenrow); 
      else fprintf (fp, "# center    : %f\n", args->cenrow + 1);
      fprintf (fp, "# peak width: %f\n", args->platw);
      fprintf (fp, "# base width: %f\n", args->basew);
    }
  else if ((args->type == 1) || (args->type == 3) || (args->type == 4))
    {
      fprintf (fp, "# type         : simple\n");
      fprintf (fp, "# min row      : %d\n", args->minrow);
      fprintf (fp, "# max row      : %d\n", args->maxrow);
    }
  else 
    {
      fprintf (fp, "# type         : all\n");
    }
  return 0;
}
  

char *
logfile_time ()
{
  time_t now;
  struct tm *mytime;
  static char timestring[26];

  now = time (NULL);
  mytime = (struct tm *) localtime (&now);
  strncpy (timestring, (char *) asctime (mytime), 26);
  timestring[24] = '\0';
  return timestring;
}


static error_t
parse_opt (int key, char *arg, struct argp_state *state)
{
/* Our structure with the argument data */
  struct arguments *arguments = state->input;

/* Temp variables to hold values from sscanf */
  int min, max;
  int total_scanned;
  double cenrow, platw, basew;

  switch (key)
    {
    case 'v':
      arguments->verbose = 1;
      break;

    case 'e':
      arguments->echelle = 1;
      break;
    
    case 'i':
      arguments->interactive = 1;
      break;

    case 'u':
      arguments->unifspec = 1;
      break;
	
	case 'f':
	  arguments->fits = 1;
	  break;

    case 'o':
      arguments->output_file = arg;
      break;

    case 't':
      if (arguments->type != 0)
        {
          fprintf (stderr,
                   "ERROR: Cannot perform more than one type of extraction at once.\n");
          argp_usage (state);
        }
      if (sscanf (arg, "%d,%d", &min, &max) != 2)
        {
          fprintf (stderr,
                   "ERROR: Need two integer columns for vertical extraction.\n");
          argp_usage (state);
        }
      else
        {
          arguments->minrow = min;
          arguments->maxrow = max;
          arguments->type = 3;
        }
      break;

    case 'l':
      arguments->logfile = arg;
      break;

    case 's':
      if (arguments->type != 0)
        {
          fprintf (stderr,
                   "ERROR: Cannot perform both simple and careful extractions!\n");
          argp_usage (state);
        }
      if (sscanf (arg, "%d,%d", &min, &max) != 2)
        {
          fprintf (stderr,
                   "ERROR: Need two integer rows for simple extraction.\n");
          argp_usage (state);
        }
      else
        {
          arguments->minrow = min;
          arguments->maxrow = max;
          arguments->type = 1;
        }
      break;

    case 'c':
      if (arguments->type != 0)
        {
          fprintf (stderr,
                   "ERROR: Cannot perform both simple and careful extractions!\n");
          argp_usage (state);
        }
      total_scanned = sscanf (arg, "%lf,%lf,%lf", &platw, &basew, &cenrow);
      if (total_scanned == 3)
        {
          arguments->guess_center = 0;
          arguments->cenrow = cenrow;
          arguments->platw = platw;
          arguments->basew = basew;
          arguments->type = 2;
        }
      else if (total_scanned == 2)
        {
          arguments->platw = platw;
          arguments->basew = basew;
          arguments->guess_center = 1;
          arguments->type = 2;
        }
      else
        {
          fprintf (stderr,
                   "ERROR: Need at least the plateau and base width for careful extraction: %s \n",
                   arg);
          argp_usage (state);
        }
      break;

    case ARGP_KEY_ARG:
      if (state->arg_num >= 1)
        argp_usage (state);
      arguments->input_file = arg;
      break;

    case ARGP_KEY_END:
      if (state->arg_num < 0)
        argp_usage (state);
      break;

    default:
      return ARGP_ERR_UNKNOWN;
    }
  return 0;
}
